/*
Stochastic Theory Education through Visualization Environment (STEVE) 2.0
Robert Florence, Tennessee Tech University
Visualization.cpp - This class creates a visualization from the output of the
simulation program and draws it using OpenGL.
Last Modified: July 24, 2008
*/

#include "Visualization.h"

/*
create a new three-dimensional map from the
    output of the simulation program
period: number of timesteps in the visualization
filename: file to read simulation program output from
*/
Visualization::Visualization(int period, char* filename)
{
    int time, row, col, temp;
    std::ifstream inFile(filename);
    
    //initialize variables
    simPeriod = period;
    frame = -1;
    
    //create three dimensional array
    map = new float**[simPeriod];
    
    for (time = 0; time < simPeriod; time++)
    {
        //create rows
        map[time] = new float*[MAP_HEIGHT];
        
        for (row = 0; row < MAP_HEIGHT; row++)
        {
            //create columns
            map[time][row] = new float[MAP_WIDTH];
        }
    }
    
    //fill three dimensional array
    for (time = 0; time < simPeriod && inFile.good(); time++)
    {
        //read timestamp (not used)
        inFile >> temp;

        //read map at current timestamp
        for (row = 0; row < MAP_HEIGHT; row++)
        {
            for (col = 0; col < MAP_WIDTH && inFile.good(); col++)
            {
                inFile >> map[time][row][col];
            }
        }
    }
    
    inFile.close();
}

/*
remove the visualization from memory
*/
Visualization::~Visualization()
{
    int time, row;
    
    //delete rows
    for (time = 0; time < simPeriod; time++)
    {
        //delete columns
        for (row = 0; row < MAP_HEIGHT; row++)
        {
            delete[] map[time][row];
        }
        
        delete[] map[time];
    }
    
    //delete time dimension
    delete[] map;
}

/*
draw the map data at the current frame
left, top: either MAP1_LEFT, MAP1_TOP or MAP2_LEFT, MAP2_TOP
*/
void Visualization::draw(float left, float top)
{
    int row, col;
    Color color1, color2, color3, color4, avgColor;
    char timePeriod[25];
    
    //avoid drawing when frame is invalid
    if (frame < 0 || frame >= simPeriod)
        return;
        
    //draw time period in center of window
    sprintf(timePeriod, "Time Period: %d hr", frame+1);
    glColor3f(0.0f, 0.0f, 0.0f);
    drawText(MAP1_LEFT + 0.037*MAP_WIDTH, MAP2_TOP + 0.17, GLUT_BITMAP_HELVETICA_12, timePeriod);
        
    //draw center of map (0.05 border around edge must be drawn separately)
    for (row = 0; row < MAP_HEIGHT - 1; row++)
    {
        for (col = 0; col < MAP_WIDTH - 1; col++)
        {
            glPushMatrix();
            //move to row and column
            glTranslatef(0.1*col + left + 0.05, -0.1*row + top - 0.05, 0.0f);
            
            //get color of four adjacent grids
            color1 = getColor(map[frame][row][col]);
            color2 = getColor(map[frame][row+1][col]);
            color3 = getColor(map[frame][row+1][col+1]);
            color4 = getColor(map[frame][row][col+1]);
            //get average color for center of four adjacent grids
            avgColor.setRGB((color1.getRed() + color2.getRed() + color3.getRed() + color4.getRed())/4,
                (color1.getGreen() + color2.getGreen() + color3.getGreen() + color4.getGreen())/4,
                (color1.getBlue() + color2.getBlue() + color3.getBlue() + color4.getBlue())/4);
                
            //draw square from (row, col) to (row+1, col+1)
            glBegin(GL_TRIANGLE_FAN);
                glColor3f(avgColor.getRed(), avgColor.getGreen(), avgColor.getBlue());
                glVertex2f(0.05f, -0.05f);
                glColor3f(color1.getRed(), color1.getGreen(), color1.getBlue());
                glVertex2f(0.0f, 0.0f);
                glColor3f(color2.getRed(), color2.getGreen(), color2.getBlue());
                glVertex2f(0.0f, -0.1f);
                glColor3f(color3.getRed(), color3.getGreen(), color3.getBlue());
                glVertex2f(0.1f, -0.1f);
                glColor3f(color4.getRed(), color4.getGreen(), color4.getBlue());
                glVertex2f(0.1f, 0.0f);
                glColor3f(color1.getRed(), color1.getGreen(), color1.getBlue());
                glVertex2f(0.0f, 0.0f);
            glEnd();

            glPopMatrix();
        }
    }
    
    //draw top and bottom edges of map
    for (row = 0; row < MAP_HEIGHT; row += MAP_HEIGHT - 1)
    {
        for (col = 0; col < MAP_WIDTH - 1; col++)
        {
            glPushMatrix();
            //move to row and column
            if (row == 0)
                glTranslatef(0.1*col + left + 0.05, -0.1*row + top, 0.0f);
            else
                glTranslatef(0.1*col + left + 0.05, -0.1*row + top - 0.05, 0.0f);

            //get color of two adjacent edge grids
            color1 = getColor(map[frame][row][col]);
            color2 = getColor(map[frame][row][col+1]);

            //draw rectangle on edge of map from (row, col) to (row, col+1)
            glBegin(GL_QUADS);
                glColor3f(color1.getRed(), color1.getGreen(), color1.getBlue());
                glVertex2f(0.0f, 0.0f);
                glVertex2f(0.0f, -0.05f);
                glColor3f(color2.getRed(), color2.getGreen(), color2.getBlue());
                glVertex2f(0.1f, -0.05f);
                glVertex2f(0.1f, 0.0f);
            glEnd();
            
            glPopMatrix();
        }
    }
    
    //draw right and left edges of map
    for (col = 0; col < MAP_WIDTH; col += MAP_WIDTH - 1)
    {
        for (row = 0; row < MAP_HEIGHT - 1; row++)
        {
            glPushMatrix();
            //move to row and column
            if (col == 0)
                glTranslatef(0.1*col + left, -0.1*row + top - 0.05, 0.0f);
            else
                glTranslatef(0.1*col + left + 0.05, -0.1*row + top - 0.05, 0.0f);

            //get color of two adjacent edge grids
            color1 = getColor(map[frame][row][col]);
            color2 = getColor(map[frame][row+1][col]);

            //draw rectangle on edge of map from (row, col) to (row+1, col)
            glBegin(GL_QUADS);
                glColor3f(color1.getRed(), color1.getGreen(), color1.getBlue());
                glVertex2f(0.0f, 0.0f);
                glVertex2f(0.05f, 0.0f);
                glColor3f(color2.getRed(), color2.getGreen(), color2.getBlue());
                glVertex2f(0.05f, -0.1f);
                glVertex2f(0.0f, -0.1f);
            glEnd();

            glPopMatrix();
        }
    }
    
    //draw four corners of map
    for (row = 0; row < MAP_HEIGHT; row += MAP_HEIGHT - 1)
    {
        for (col = 0; col < MAP_WIDTH; col += MAP_WIDTH - 1)
        {
            glPushMatrix();
            //move to row and column
            glTranslatef(0.1*col + left, -0.1*row + top, 0.0f);
            if (row > 0)
                glTranslatef(0.0f, -0.05f, 0.0f);
            if (col > 0)
                glTranslatef(0.05f, 0.0f, 0.0f);
            
            //get color of corner grid
            color1 = getColor(map[frame][row][col]);
            
            //draw square on corner of map at (row, col)
            glBegin(GL_QUADS);
                glColor3f(color1.getRed(), color1.getGreen(), color1.getBlue());
                glVertex2f(0.0f, 0.0f);
                glVertex2f(0.05f, 0.0f);
                glVertex2f(0.05f, -0.05f);
                glVertex2f(0.0f, -0.05f);
            glEnd();

            glPopMatrix();
        }
    }

}

/*
draw the map title, axes, and grid
left, top: either MAP1_LEFT, MAP1_TOP or MAP2_LEFT, MAP2_TOP
*/
void Visualization::drawAxes(float left, float top)
{
    int i;
    
    glPushMatrix();
    glTranslatef(left, top, 0.0f);
    glColor3f(0.0f, 0.0f, 0.0f);
    
    //draw map title
    if (top == MAP1_TOP)
        drawText(0.041*MAP_WIDTH, 0.05f, GLUT_BITMAP_HELVETICA_12, "Default Map");
    else if (top == MAP2_TOP)
        drawText(0.041*MAP_WIDTH, 0.05f, GLUT_BITMAP_HELVETICA_12, "Custom Map");
    
    //draw y axis labels
    drawText(-0.24f, -0.02f, GLUT_BITMAP_HELVETICA_10, "300 km");
    drawText(-0.24f, -0.42f, GLUT_BITMAP_HELVETICA_10, "200 km");
    drawText(-0.24f, -0.82f, GLUT_BITMAP_HELVETICA_10, "100 km");
    drawText(-0.175f, -1.22f, GLUT_BITMAP_HELVETICA_10, "0 km");
    
    //draw x axis labels
    drawText(-0.05f, -0.1*MAP_HEIGHT - 0.1, GLUT_BITMAP_HELVETICA_10, "0 km");
    drawText(0.32f, -0.1*MAP_HEIGHT - 0.1, GLUT_BITMAP_HELVETICA_10, "100 km");
    drawText(0.72f, -0.1*MAP_HEIGHT - 0.1, GLUT_BITMAP_HELVETICA_10, "200 km");
    drawText(1.12f, -0.1*MAP_HEIGHT - 0.1, GLUT_BITMAP_HELVETICA_10, "300 km");
    drawText(1.52f, -0.1*MAP_HEIGHT - 0.1, GLUT_BITMAP_HELVETICA_10, "400 km");
    drawText(1.92f, -0.1*MAP_HEIGHT - 0.1, GLUT_BITMAP_HELVETICA_10, "500 km");

    //draw y grid lines
    for (i = 0; i <= MAP_HEIGHT; i++)
    {
        //larger mark every 4 rows
        if (i % 4 == 0)
        {
            glBegin(GL_LINES);
                glVertex2f(-0.04f, 0.0f);
                glVertex2f(0.1*MAP_WIDTH, 0.0f);
            glEnd();
        }
        else
        {
            glBegin(GL_LINES);
                glVertex2f(-0.02f, 0.0f);
                glVertex2f(0.1*MAP_WIDTH, 0.0f);
            glEnd();
        }
        glTranslatef(0.0f, -0.1f, 0.0f);
    }
    glTranslatef(0.0f, 0.1f, 0.0f);
    
    //draw x grid lines
    for (i = 0; i <= MAP_WIDTH; i++)
    {
        //larger mark every 4 columns
        if (i % 4 == 0)
        {
            glBegin(GL_LINES);
                glVertex2f(0.0f, -0.04f);
                glVertex2f(0.0f, 0.1*MAP_HEIGHT);
            glEnd();
        }
        else
        {
            glBegin(GL_LINES);
                glVertex2f(0.0f, -0.02f);
                glVertex2f(0.0f, 0.1*MAP_HEIGHT);
            glEnd();
        }
        glTranslatef(0.1f, 0.0f, 0.0f);
    }
    
    glPopMatrix();
}

/*
draw the color bar on the right of the window
*/
void Visualization::drawColorBar()
{
    int i;
    float interval;
    char intervalStr[10];

    //divide color bar size into 5 sections
    interval = (-0.2*MAP_HEIGHT - 0.4)/5;

    glPushMatrix();
    glTranslatef(MAP1_LEFT + 0.1*MAP_WIDTH + 0.2, MAP1_TOP, 0.0f);

    //draw color bar
    glBegin(GL_QUAD_STRIP);
        //50 mm/hr = red
        glColor3f(1.0f, 0.0f, 0.0f);
        glVertex2f(0.0f, 0.0f);
        glVertex2f(0.3f, 0.0f);
        //40 mm/hr = yellow
        glColor3f(1.0f, 1.0f, 0.0f);
        glVertex2f(0.0f, interval);
        glVertex2f(0.3f, interval);
        //30 mm/hr = green
        glColor3f(0.0f, 1.0f, 0.0f);
        glVertex2f(0.0f, 2*interval);
        glVertex2f(0.3f, 2*interval);
        //20 mm/hr = cyan
        glColor3f(0.0f, 1.0f, 1.0f);
        glVertex2f(0.0f, 3*interval);
        glVertex2f(0.3f, 3*interval);
        //10 mm/hr = blue
        glColor3f(0.0f, 0.0f, 1.0f);
        glVertex2f(0.0f, 4*interval);
        glVertex2f(0.3f, 4*interval);
        //0 mm/hr = white
        glColor3f(1.0f, 1.0f, 1.0f);
        glVertex2f(0.0f, 5*interval);
        glVertex2f(0.3f, 5*interval);
    glEnd();

    //draw color bar frame
    glColor3f(0.0f, 0.0f, 0.0f);
    glBegin(GL_LINE_LOOP);
        glVertex2f(0.0f, 0.0f);
        glVertex2f(0.3f, 0.0f);
        glVertex2f(0.3f, 5*interval);
        glVertex2f(0.0f, 5*interval);
    glEnd();

    //draw bar labels
    for (i = 0; i <= 5; i++)
    {
        glBegin(GL_LINES);
            glVertex2f(0.3f, i*interval);
            glVertex2f(0.34f, i*interval);
        glEnd();

        //create interval label
        if (i == 0)
            sprintf(intervalStr, "%d+ mm/hr", 10*(5-i));
        else
            sprintf(intervalStr, "%d mm/hr", 10*(5-i));

        drawText(0.36f, i*interval - 0.02, GLUT_BITMAP_HELVETICA_12, intervalStr);
    }

    glPopMatrix();
}

/*
move to previous or next frame
*/
void Visualization::decFrame()
{
    frame--;
    
    if (frame < 0)
        frame = simPeriod - 1;
}

void Visualization::incFrame()
{
    frame++;
    
    if (frame >= simPeriod)
        frame = 0;
}

/*
determine the color of the visualization map of the specified output value
val: value of a grid on the map
return: color for the value according to the color bar
*/
Color Visualization::getColor(float val)
{
    Color color;

    //0.0 = white, 10.0 = blue
    if (val > 0.0 && val <= 10.0)
        color.setRGB(1.0 - val/10.0, 1.0 - val/10.0, 1.0f);
    //10.0 = blue, 20.0 = cyan
    else if (val > 10.0 && val <= 20.0)
        color.setRGB(0.0f, (val - 10.0)/10.0, 1.0f);
    //20.0 = cyan, 30.0 = green
    else if (val > 20.0 && val <= 30.0)
        color.setRGB(0.0f, 1.0f, 1.0 - (val - 20.0)/10.0);
    //30.0 = green, 40.0 = yellow
    else if (val > 30.0 && val <= 40.0)
        color.setRGB((val - 30.0)/10.0, 1.0f, 0.0f);
    //40.0 = yellow, 50.0 = red
    else if (val > 40.0 && val <= 50.0)
        color.setRGB(1.0f, 1.0 - (val - 40.0)/10.0, 0.0f);
    //50.0+ = red
    else if (val > 50.0)
        color.setRGB(1.0f, 0.0f, 0.0f);

    return color;
}

/*
draw text to the window
x, y: GL coordinates to draw the text
font: GLUT font to draw the text in
text: string to draw
*/
void Visualization::drawText(float x, float y, void* font, char* text)
{
    char* c;
    
    glRasterPos2f(x, y);
    
    //draw each character
    for (c = text; *c != '\0'; c++)
    {
        glutBitmapCharacter(font, *c);
    }
}
